import os
import datetime
import json
import logging
from subprocess import Popen, PIPE
from threading import Thread

# {CATTURA_DIR}/bin/apis/python
CATTURA_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
CATTURA_LOG_DIR = os.path.join(CATTURA_DIR,'logs')

def __implement():
    loggers = {}
    LOG_DATE_FORMAT = "%Y-%m-%d.%I-%M-%S-%p"
    LOG_RECORD_FORMAT = "%(asctime)s [%(name)s] %(levelname)s: %(message)s"
    GRAPH_JSON = "graph.json"

    def getLogger(name):
        if not name in loggers:
            logger = logging.getLogger(name)
            logdir = os.path.join(CATTURA_LOG_DIR, name)
            if not os.path.exists(logdir):
                os.makedirs(logdir)
            logfilename = os.path.join(logdir,"{date}.log".format(date=datetime.datetime.today().strftime(LOG_DATE_FORMAT)))
            
            # handlers
            logFileHandler = logging.FileHandler(logfilename)
            logStreamHandler = logging.StreamHandler()

            # formatter
            logFormatter = logging.Formatter(LOG_RECORD_FORMAT)
            logFileHandler.setFormatter(logFormatter)
            logStreamHandler.setFormatter(logFormatter)

            logger.addHandler(logStreamHandler)
            logger.addHandler(logFileHandler)
            loggers[name] = logger
        return loggers[name]

    class MediaProcessingAgent(object):
        def __init__(self,path=None,quiet=False):
            if path is None:
                path = "media-processing-agent-1.0.0.jar"
            self.path = path
            self.quiet = quiet
            self.process = None
            self.listeners = []

        def on(self,event):
            def register(func):
                if hasattr(event,'__call__'):
                    def wrapped(notification):
                        if event(notification):
                            func(notification)
                else:
                    def wrapped(notification):
                        if notification['notification'] == event:
                            func(notification)
                self.listeners.append(wrapped)
            return register

        def onMessage(self,notification):
            for listener in self.listeners:
                listener(notification)

        def sendNotification(self,name,body=None):
            if self.process is None:
                return

            self.process.stdin.write(json.dumps({
                "notification": name,
                "body": body
            }))
            self.process.stdin.write("\n")
            self.process.stdin.flush()

        def start(self):
            self.process = Popen(['java','-jar',self.path],stdin=PIPE,stdout=PIPE,stderr=PIPE if self.quiet else None)
            def listener():
                while True:
                    try:
                        line = self.process.stdout.readline()
                        try:
                            note = json.loads(line)
                            self.onMessage(note)
                        except ValueError:
                            if not self.quiet:
                                print line
                    except KeyboardInterrupt: break
            Thread(target=listener).start()

        def kill(self):
            if self.process is None:
                return

            self.process.kill()

    class Template(object):
        def __init__ (self, path):
            self.path = path
            self.id = os.path.basename(path)
            self.graphPath = os.path.join(self.path, GRAPH_JSON)
            with open(self.graphPath) as graph:
                self.graph = json.loads(graph.read())
                for node in self.graph['nodes']:
                    if isinstance(node['config'], basestring):
                        node['config'] = json.loads(node['config'])

        def serialize (self):
            nodes = []
            for node in self.graph['nodes']:
                newNode = {}
                newNode.update(node)
                if not isinstance(newNode['config'], basestring):
                    newNode['config'] = json.dumps(newNode['config'])
                nodes.append(newNode)
            return json.dumps({"nodes": nodes})

        def nodes (self):
            for node in self.graph['nodes']:
                yield Node(self, node)

        def save (self):
            with open(self.graphPath, 'w') as graph:
                graph.write(self.serialize())
    
    class Node(object):
        def __init__(self, template, node):
            object.__setattr__(self, "node", node)
            object.__setattr__(self, "template", template)
        
        def __getattr__(self, name):
            return self.node[name]
        
        def __setattr__(self, name, value):
            self.node[name] = value

    def templates ():
        templateDir = os.path.join(CATTURA_DIR, "storage", "templates")
        for item in os.listdir(templateDir):
            try:
                yield Template(os.path.join(templateDir, item))
            except Exception as e: print e

    return templates, getLogger, MediaProcessingAgent

templates, getLogger, MediaProcessingAgent = __implement()